import os
import shutil
import logging
from laboro.error import LaboroError


class Workspace:
  """The ``laboro.workspace.Workspace`` object manages the workflow and instances workspaces.

  The workflow workspace is a directory named after the workflow name within the Laboro global temp directory.

  Instances workspaces are a workflow workspace sub-directory named after the workflow session.

  A ``Workspace`` instance is used to store files and data within the workflow execution.

  Arguments:
    workspacedir: A string defining the directory where all workspaces are stored.
    workflow: A string defining the workflow name.
    session: A unique string defining the workflow execution session.

  Returns:
    ``laboro.workspace.Workspace``: The workspace for the workflow session.

  Raises:
    ``laboro.error.LaboroError``: When any error such as *OSError* while creating the workspace occurs.
  """
  def __init__(self, workspacedir, workflow, session):
    try:
      self.workspacepath = os.path.join(workspacedir, workflow, session)
      logging.info(f"[+] Creating workspace: {self.workspacepath}")
      os.makedirs(self.workspacepath)
    except Exception as err:
      raise LaboroError(f"[{err.__class__.__name__}] {str(err)}") from err

  def delete(self):
    """Deletes the instance workspace and all its content.
    This will **not** delete the workflow workspace.

    Raises:
      ``laboro.error.LaboroError``: When any error such as *OSError* while creating the workspace occurs.
    """
    logging.info(f"[+] Deleting workspace: {self.workspacepath}")
    try:
      shutil.rmtree(self.workspacepath)
    except Exception as err:
      raise LaboroError(f"[{err.__class__.__name__}] {str(err)}") from err



import os
import logging
from laboro.error import LaboroError
from laboro.history import History
from laboro.workspace import Workspace
from laboro.config.manager import Manager as ConfigMgr
from laboro.module.manager import Manager as ModuleMgr
from laboro.error.handler import Handler as ErrorHandler


class Context:
  def __init__(self, logger):
    self.logger = logger
    self.modules = dict()
    self.history = None
    self.modulemgr = ModuleMgr()
    self.configmgr = ConfigMgr("/etc/laboro/laboro.yml")
    self.workspacedir = self.configmgr.get_parameter("main",
                                                     "$.laboro.workspacedir")
    self.histdir = self.configmgr.get_parameter("main",
                                                "$.laboro.histdir")
    self.workflowdir = self.configmgr.get_parameter("main",
                                                    "$.laboro.workflowdir")
    self.logdir = self.configmgr.get_parameter("main",
                                               "$.laboro.log.dir")
    self.loglevel = self.configmgr.get_parameter("main",
                                                 "$.laboro.log.level")
    self.workspace = None
    logging.getLogger().vault.clear()

  def reset(self, workflow_name, session):
    logging.getLogger().log_section("WORKFLOW",
                                    f"Started {workflow_name} / {session}")
    logging.getLogger().vault.clear()
    self._configure_logger(workflow_name, session)
    self.set_workflow_config(workflow_name)
    self.workspace = Workspace(workspacedir=self.workspacedir,
                               workflow=workflow_name,
                               session=session)
    self.history = History(filename=os.path.join(self.histdir,
                                                 f"{workflow_name}.db"),
                           workflow=workflow_name,
                           session=session,
                           params=self.configmgr.workflow_config)
    self.history.enter()
    self.install_packages()
    return self

  def exit(self, session, kind, value, traceback, exit_code):
    logging.info(f"[+] Workflow ended with code: {exit_code}")
    self.workspace.delete()
    self.history.exit(kind, value)
    self.handle_error(kind, value, traceback)
    self.logger.remove_file_handler(session)

  def _configure_logger(self, workflow_name, session):
    self.logger.add_file_handler(self.logdir, workflow_name, session)
    self.logger.set_log_level(self.loglevel)

  def set_workflow_config(self, workflow_name):
    config_file_name = os.path.join(self.workflowdir, f"{workflow_name}.yml")
    logging.info(f"[+] Loading workflow_config: {config_file_name}")
    self.configmgr.workflow_config = config_file_name

  def install_packages(self):
    """Install all packages listed in the workflow configuration file.
    """
    packages = self.configmgr.get_parameter("workflow",
                                            "$.laboro.workflow.packages")
    for pkg in packages:
      self.modulemgr.install_package(pkg)

  def register_class(self, module, cls):
    self.modulemgr.register_class_from_module(cls, module)

  def register_instance_secrets(self, instance, args):
    class_args = instance.specification["args"]
    secret_keys = [arg["name"] for arg in class_args if arg["secret"]]
    secrets = [args[key] for key in args.keys() if key in secret_keys]
    list(map(logging.getLogger().vault.add, secrets))

  def register_method_secrets(self, instance, action, args):
    try:
      method_args = [meth for meth in instance.specification["methods"] if meth["name"] == action][0]["args"]
      secret_keys = [arg["name"] for arg in method_args if arg["secret"]]
      secrets = [args[key] for key in args.keys() if key in secret_keys]
      list(map(logging.getLogger().vault.add, secrets))
    except IndexError as err:
      raise LaboroError(f"UnknownMethodError: Unknown method {instance.__class__.__name__}.{action}") from err

  def handle_error(self, kind, value, traceback):
    ErrorHandler().handle_error(kind, value, traceback)



class LaboroError(Exception):
  """Laboro specific error class.

  Arguments:
    message: A string defining the message to be displayed when error occurs.

  Returns:
    ``laboro.error.LaboroError``: A Laboro Error.

  """
  def __init__(self, message):
    self.message = message
    super().__init__(self.message)



import logging


class LaboroLogger(logging.Logger):
  """The ``laboro.logger.LaboroLogger`` overrides the ``logging.Logger`` class.

  It adds a ``laboro.vault.Vault`` instance to the logger object allowing secret management.

  Any secret stored in the vault instance will be redacted before logging.
  """

  @staticmethod
  def log_line():
    logging.info("-" * 80)

  @staticmethod
  def log_section(prefix, message):
    LaboroLogger.log_line()
    logging.info(f"[{prefix}] {message}")
    LaboroLogger.log_line()

  def __init__(self, name, level=logging.NOTSET):
    self.vault = None
    super(LaboroLogger, self).__init__(name, level)

  def set_vault(self, vault):
    """Set the ``laboro.vault.Vault`` instance used to prevent secrets to be logged."""
    self.vault = vault

  def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
    """Override logging.Logger.makeRecord.
    """
    msg = self.vault.protect(msg)
    return super(LaboroLogger, self).makeRecord(name, level, fn, lno, msg, args, exc_info, func, extra, sinfo)



import sys
import logging
import traceback


class Handler:
  """The ``laboro.error.handler.Handler`` is a singleton that logs every errors given to its ``handle_error()`` method as critical then exit with code 1.

  This object is useful with object with runtime context.

  ..  code-block:: python

      from laboro.error.handler import Handler

      class FakeObj:
        def __enter__(self):
          pass

        def __exit__(self, kind, value, trace):
          Handler.handle_error(kind, value, trace)

      with FakeObj() as obj:
        raise Exception("This Exception will be logged as critical and backtracked.")
  """

  @staticmethod
  def handle_error(kind, value, trace):
    """Log the given error with its traceback and exit with exit code 1."""
    if kind is not None:
      logging.critical(f"[{kind.__name__}]: {str(value)}")
      if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
        traceback.print_tb(trace)
      if kind != SystemExit:
        sys.exit(1)



import os
import logging
from laboro.error import LaboroError
from laboro.validator import Validator


class Module:
  def __init__(self, filepath, args=None):
    self.filepath = filepath
    self.args = args
    self.specification = self._validate_spec()
    self._validate_args()

  @staticmethod
  def laboro_method(func):
    def wrapper(self, *args, **kwargs):
      if len(args) > 1:
        logging.error("Using method with args instead of kwargs !")
        logging.warning("No argument checking will be done.")
        func(*args[1:])
      else:
        try:
          method = func.__name__
          methods = self.specification["methods"]
          method = [meth for meth in methods if meth["name"] == method][0]
          Validator().validate_method_args(method, kwargs)
        except IndexError as err:
          raise LaboroError(f"UnknownMethodError: Unknown method: {method}") from err
        func(self, **kwargs)
    return wrapper

  def _validate_spec(self):
    """Load module specification from its YAML data file."""
    base_spec = os.path.join(os.path.dirname(__file__),
                             "schema",
                             "object.yml")
    spec_file = os.path.join(os.path.dirname(self.filepath),
                             "schema",
                             "specification.yml")
    return Validator().validate_instance(schema=base_spec,
                                         instance=spec_file)

  def _validate_args(self):
    """Validate arguments against a specification.
    """
    spec_file = os.path.join(os.path.dirname(self.filepath),
                             "schema",
                             "specification.yml")
    return Validator().validate_obj_args(specification=spec_file,
                                         args=self.args)

  def get_arg_value(self, arg):
    """Get the value of the specified argument.

    Arguments:
      arg: The argument to get the value from.

    Returns:
      The value of the searched argument. The type of the return value depend of the value of the searched argument.

    Raises:
      ``laboro.error.LaboroError``: When the specified argument is not known to the module.
    """
    if arg in self.args:
      return self.args[arg]
    msg = {"type": "UnknownModuleArgError",
           "message": f"Module argument not found: {arg}"}
    raise LaboroError(f"[{msg['type']}] {msg['message']}")

  def get_arg_value_as_string(self, arg):
    """Get the value of the specified argument as a string.

    Arguments:
      arg: The argument to get the value from.

    Returns:
      ``str``: The string representation of value of the searched argument.

    Raises:
      ``laboro.error.LaboroError``: When the specified argument is not known to the module.
    """
    if arg in self.args:
      return str(self.args[arg])
    msg = {"type": "UnknownModuleArgError",
           "message": f"Module argument not found: {arg}"}
    raise LaboroError(f"[{msg['type']}] {msg['message']}")



import yaml
import logging
import builtins
from jsonschema import validate
from jsonschema.exceptions import ValidationError, SchemaError
from laboro.error import LaboroError


class Validator:
  """The ``laboro.Validator`` object purpose is to validate specification and arguments.

  This object is instantiated by some **Laboro** core components and is not intended to be used directly.

  Arguments:
    ``filepath``: A string specifying the filepath of **the base module**. The expected value is the base module ``__file__`` attribute value. The specification will be search as ``object.yml`` in the ``specification`` directory situated at the same level of ``filepath``.

  Raises:
    ``laboro.error.Error``: When the **Laboro** base module specification file is not found or not a valid *YAML* file.
  """

  def _load_file(self, filepath):
    try:
      with open(filepath, mode="r", encoding="utf-8") as schema:
        return yaml.load(schema, Loader=yaml.FullLoader)
    except FileNotFoundError as err:
      msg = {"type": "FileNotFoundError",
             "message": f"Schema file not found: {filepath}"}
      logging.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err
    except yaml.scanner.ScannerError as err:
      msg = {"type": "InvalidYamlError",
             "message": f"YAML file is invalid: {filepath}"}
      logging.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err

  def validate_instance(self, schema, instance):
    """Validate the given instance against the given schema.

    Arguments:
    ``schema``: A file path to a *YAML* representation of a `jsonschema` representing the schema against which the ``instance`` should be validated.

    ``instance``: A file path to a *YAML* representation of a `jsonschema` representing the instance to validate.

    Raises:
      ``laboro.error.Error``: Whenever the schema or the instance file can't be loaded or are not valid *YAML* files. Error is also raised when the given ``instance`` is not valid.
    """
    try:
      schema_yml = self._load_file(schema)
      instance_yml = self._load_file(instance)
      validate(instance=instance_yml, schema=schema_yml)
      return instance_yml
    except ValidationError as err:
      err_msg = str(err).split("\n")[0]
      msg = {"type": "InvalidInstanceError",
             "message": f"Instance is invalid: {err_msg}"}
      logging.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err
    except SchemaError as err:
      msg = {"type": "InvalidSchemaError",
             "message": f"Schema is invalid: {err}"}
      logging.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err

  def validate_data(self, specification, data):
    try:
      specification_yml = self._load_file(specification)
      validate(instance=data, schema=specification_yml)
    except ValidationError as err:
      err_msg = str(err).split("\n")[0]
      msg = f"[InvalidDataError] {err_msg}"
      logging.critical(msg)
      raise LaboroError(msg) from err

  def validate_method_args(self, method, args):
    try:
      validate(instance=args, schema=method)
      if isinstance(args, dict):
        if self._validate_required(args, method):
          for arg, value in args.items():
            if arg in [arg["name"] for arg in method["args"]]:
              self._validate_type(arg, value, method)
              self._validate_implied(arg, args, method)
              self._validate_excluded(arg, args, method)
            else:
              msg = {"type": "UnknownArgError",
                     "message": f"Unknown argument: {arg}"}
              logging.critical(f"[{msg['type']}] {msg['message']}")
              raise LaboroError(f"[{msg['type']}] {msg['message']}")
          return True
        msg = "[MissingRequiredArgError] Missing required argument."
        logging.critical(msg)
        raise LaboroError(msg)
      msg = f"[ModuleArgError]: Expected a dict as args, received {type(args)}"
      logging.critical(msg)
      raise LaboroError(msg)
    except ValidationError as err:
      err_msg = str(err).split("\n")[0]
      msg = f"[InvalidSpecError] {err_msg}"
      logging.critical(msg)
      raise LaboroError(msg) from err

  def validate_obj_args(self, specification, args):
    try:
      specification_yml = self._load_file(specification)
      validate(instance=args, schema=specification_yml)
      if isinstance(args, dict):
        if self._validate_required(args, specification_yml):
          for arg, value in args.items():
            if arg in [arg["name"] for arg in specification_yml["args"]]:
              self._validate_type(arg, value, specification_yml)
              self._validate_implied(arg, args, specification_yml)
              self._validate_excluded(arg, args, specification_yml)
            else:
              msg = {"type": "UnknownArgError",
                     "message": f"Unknown argument: {arg}"}
              logging.critical(f"[{msg['type']}] {msg['message']}")
              raise LaboroError(f"[{msg['type']}] {msg['message']}")
          return True
        msg = "[MissingRequiredArgError] Missing required argument."
        logging.critical(msg)
        raise LaboroError(msg)
      msg = f"[ModuleArgError]: Expected a dict as args, received {type(args)}"
      logging.critical(msg)
      raise LaboroError(msg)
    except ValidationError as err:
      err_msg = str(err).split("\n")[0]
      msg = f"[InvalidSpecError] {err_msg}"
      logging.critical(msg)
      raise LaboroError(msg) from err

  def _validate_required(self, arguments, specification):
    required = sorted([arg["name"] for arg in list(filter(lambda arg: arg["required"], specification["args"]))])
    return sorted([arg for arg in arguments if arg in required]) == required

  def _validate_type(self, arg, value, specification):
    arg_type = [aarg["type"] for aarg in specification["args"] if aarg["name"] == arg][0]
    try:
      arg_type = getattr(builtins, arg_type)
      if not isinstance(value, arg_type):
        msg = {"type": "BadArgTypeError",
               "message": f"Expected type {arg_type} / received {type(value)} for {arg}."}
        logging.critical(f"[{msg['type']}] {msg['message']}")
        raise LaboroError(f"[{msg['type']}] {msg['message']}")
    except AttributeError as err:
      msg = {"type": "UnknownTypeError",
             "message": f"Unknown type '{arg_type}' for argument {arg}."}
      logging.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err

  def _validate_implied(self, arg, arguments, specification):
    try:
      implied = sorted([aarg["implied"] for aarg in list(filter(lambda arg: arg["implied"], specification["args"])) if aarg["name"] == arg])[0]
    except IndexError:
      implied = list()
    if sorted([iarg for iarg in arguments if iarg in implied]) != implied:
      msg = {"type": "MissingImpliedArgError",
             "message": f"All args of {implied} are required by {arg}."}
      logging.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}")

  def _validate_excluded(self, arg, arguments, specification):
    try:
      excluded = sorted([aarg["excluded"] for aarg in list(filter(lambda arg: arg["excluded"], specification["args"])) if aarg["name"] == arg])[0]
    except IndexError:
      excluded = list()
    if len([earg for earg in arguments if earg in excluded]) > 0:
      msg = {"type": "ExcludedArgError",
             "message": f"None of {' or '.join(excluded)} can be used alongside {arg}."}
      logging.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}")



import os
import logging
from jsonpath import JSONPath
from laboro.error import LaboroError
from laboro.validator import Validator


class Manager:
  """The ``laboro.config.manager.Manager`` object loads all configuration files needed by ``Laboro`` to operate.
  It first loads and checks the main configuration file then loads the workflow configuration file and checks that all parameters comply to each module definition.
  It is used by the ``workflow.manager`` to retrieve any configuration parameter needed by any module used in the workflow.

  Arguments:
    main_config: A string representing the main Laboro configuration filename.

  Returns:
    ``laboro.config.manager.Manager``: The configuration manager.

  Raises:
    ``laboro.error.LaboroError``: When one of the configuration files are not found.
  """

  @property
  def main_config(self):
    """Get the main_config as a dict."""
    return self._main_config

  @property
  def workflow_config(self):
    """Get the workflow_config as a dict."""
    return self._workflow_config

  @main_config.setter
  def main_config(self, filename):
    """Set the main_config from a YAML file.

    Arguments:
      filename: A string specifying the YAML file to load the configuration from."""
    schema = os.path.join(os.path.dirname(__file__),
                          "schema",
                          "main.yml")
    self._main_config = Validator().validate_instance(schema=schema,
                                                      instance=filename)

  @workflow_config.setter
  def workflow_config(self, filename):
    """Set the workflow_config from a YAML file.

    Arguments:
      filename: A string specifying the YAML file to load the configuration from."""
    schema = os.path.join(os.path.dirname(__file__),
                          "schema",
                          "workflow.yml")
    self._workflow_config = Validator().validate_instance(schema=schema,
                                                          instance=filename)

  def __init__(self, main_config):
    self.main_config = main_config
    self._workflow_config = dict()

  def get_parameter(self, level, param):
    """Returns the configuration parameter specified by ``param`` from the configuration ``level``.
    ``level`` is a string defining the configuration parameter level. It must be one of ``main`` or ``workflow``.

    Arguments:
      level: A string that specify the level from which retrieve the parameter.
             ``level`` Must be one of ``main`` or ``workflow``.
      param: A string that specify the parameter name as described by the JSONPath spec.

    Returns:
      Any type, depending on YAML file from which the configuration is generated.

    Raises:
      laboro.error.LaboroError: When the specified parameter is unknown or ``level`` value is not valid.

    ..  code-block:: python

        manager = Manager(main_conf, workflow_conf)
        manager.get_parameter("workflow", "$.laboro.workflow.steps[0].actions[0].input.value")
    """
    levels = {"main": self.main_config,
              "workflow": self.workflow_config}
    try:
      config = levels[level]
      try:
        return JSONPath(param).parse(config)[0]
      except IndexError as err:
        logging.critical(f"UnknownParameter: [{level}]: {param}")
        raise LaboroError(f"UnknownParameter: [{level}]: {param}") from err
    except KeyError as err:
      logging.critical(f"BadConfLevel: Bad configuration level: {level}")
      raise LaboroError(f"BadConfLevel: Bad configuration level: {level}") from err
